"""
bp_broker API module.

Listens and responds to ssl connections.  Proxies connection requests to registered event handlers.
"""


import threading
import Queue
import json
import BaseHTTPServer, SimpleHTTPServer
import ssl
import time
import re
import os
from urlparse import urlparse, parse_qs, parse_qsl
from StringIO import StringIO

import bpbroker



default_config = {
	'listen_port': 20443,
	'listen_ip': '',	# default bind to all interfaces
	'ssl_cert': '%s/dummy_api.crt' % os.path.dirname(bpbroker.__file__),
	'ssl_key': '%s/dummy_api.key' % os.path.dirname(bpbroker.__file__),
}



#####################################################

class TimeoutBaseHTTPServer(BaseHTTPServer.HTTPServer):
	timeout = 1



class APIHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

	# Response variables populated by individual request handlers
	error = 200
	error_message = ''
	content_type = "Application/json"
	data = ''


	def log_error(self, format, *args):
		pass
	def log_request(self, code='-', size='-'):
		pass


	def RequestingHost(self):  return(self.client_address[:2][0])


	def _ParseRequest(self):
		self.path = re.sub("/\?","?",self.path)
		(undef,self.package,method) = urlparse(self.path).path.split("/",2)
		self.method = re.sub("/$","",method)
		self.qs = dict(parse_qsl(urlparse(self.path).query))	# Read Get qs
		if not self.qs:	
			length = int(self.headers['Content-Length'])
			self.qs = dict(parse_qsl(self.rfile.read(length),keep_blank_values=True))


	def _ValidateRequest(self):
		error = False
		with bpbroker.config.rlock:
			if self.package not in bpbroker.config.data:  error = "Unauthorized package"
			elif '_access_key' in bpbroker.config.data[self.package] and bpbroker.config.data[self.package]['_access_key'] and \
			   ('access_key' not in self.qs or self.qs['access_key'] != bpbroker.config.data[self.package]['_access_key']):
			   	error = "Incorrect access key"
			elif re.match("_",self.method):  error = "Unauthorized method"
			else:
				try:
					self.package_obj = __import__("bpbroker."+self.package)
					for mod in self.package.split("."):  self.package_obj = getattr(self.package_obj,mod)
					if not hasattr(self.package_obj, self.method):  error = "Unauthorized method"
				except:
					try:
						self.package_obj = __import__(self.package)
						if re.search("\.",self.package):
							for mod in re.sub(".*?\.","",self.package).split("."):  self.package_obj = getattr(self.package_obj,mod)
						if not hasattr(self.package_obj, self.method):  error = "Unauthorized method"
					except:
						error = "Unauthorized method"

		if error: self.send_error(401, error)

		return(not error)


	def do_GET(self):  self.ProcessRequest()
	def do_POST(self):  self.ProcessRequest()
	def do_DELETE(self):  self.ProcessRequest()


	def ProcessRequest(self):
		self._ParseRequest()
		if self._ValidateRequest():  
			getattr(self.package_obj, self.method)(self)
			if self.error==200:  
				self.send_response(self.error)
				self.send_header('Content-Type',self.content_type)
				self.end_headers()
				self.wfile.write(self.data)
			else:  self.send_error(self.error,self.error_message)



class APIThread(threading.Thread):

	def __init__(self,worker_queue,health_queue,config={}):
		threading.Thread.__init__(self)
		self.worker_queue = worker_queue
		self.health_queue = health_queue
		self._stop_event = threading.Event()
		self.config = dict(list(default_config.items()) + list(config.items()))


	def join(self,timeout=None):
		self._stop_event.set()
		threading.Thread.join(self, timeout)


	def run(self):
		self.api_server = TimeoutBaseHTTPServer((self.config['listen_ip'], self.config['listen_port']), APIHTTPRequestHandler)
		self.api_server.socket = ssl.wrap_socket (self.api_server.socket, 
									 			  server_side=True,
									 			  certfile=self.config['ssl_cert'],
									 			  keyfile=self.config['ssl_key'],
												  )
		while not self._stop_event.is_set():
			self.api_server.handle_request()
			self.HealthCheck()


	def HealthCheck(self):
		self.health_queue.put_nowait({'thread': 'API', 'ts': int(time.time())})






"""Command-line interface to bpclient."""

import argparse
import ConfigParser
import os
import sys
import bpclient


class Args:

	def __init__(self):
		bpclient.args = self
		self.ParseArgs()
		#self.ImportIni()
		self.MergeEnvironment()
		self.MergeCommands()


	def ParseArgs(self):
		parser = argparse.ArgumentParser(description="bpclient service client")
		parser_sp1 = parser.add_subparsers(title='Commands',dest='command')

		########## Ping ###########
		parser_ping = parser_sp1.add_parser('ping', help='Connectivity health check')
		parser_ping.add_argument('--data', help='Data payload.  This is echoed back from the server on successful test')
		parser_ping.add_argument('--raw', action="store_true", default=False, help='Return raw data')


		########## Services ###########
		parser_services = parser_sp1.add_parser('service', help='Service broker registration and querying')
		parser_sp3 = parser_services.add_subparsers(dest='sub_command')

		## Register
		parser_services_register = parser_sp3.add_parser('register', help='Register supplied data to supplied key')
		parser_services_register.add_argument('--name', required=True, help='Unique key for service broker registration')
		parser_services_register.add_argument('--data', required=True, help='Data associated with this key')
		parser_services_register.add_argument('--raw', action="store_true", default=False, help='Return raw data')

		## Replace
		parser_services_replace = parser_sp3.add_parser('replace', help='Replace data associated with supplied key')
		parser_services_replace.add_argument('--name', required=True, help='Unique key')
		parser_services_replace.add_argument('--data', required=True, help='Data associated with this key')
		parser_services_replace.add_argument('--raw', action="store_true", default=False, help='Return raw data')

		## Update
		parser_services_update = parser_sp3.add_parser('update', help='Update data associated with supplied key')
		parser_services_update.add_argument('--name', required=True, help='Unique key')
		parser_services_update.add_argument('--data', required=True, help='Data associated with this key')
		parser_services_update.add_argument('--raw', action="store_true", default=False, help='Return raw data')

		## Get
		parser_services_get = parser_sp3.add_parser('get', help='Return data associated with supplied key')
		parser_services_get.add_argument('--name', required=True, help='Unique key')
		parser_services_get.add_argument('--raw', action="store_true", default=False, help='Return raw data')

		## Delete
		parser_services_delete = parser_sp3.add_parser('delete', help='Delete key from service broker')
		parser_services_delete.add_argument('--name', required=True, help='Unique key')
		parser_services_delete.add_argument('--raw', action="store_true", default=False, help='Return raw data')


		########## Execute ###########
		parser_execute = parser_sp1.add_parser('execute', help='Execute custom RPC on BP Broker server')
		parser_execute.add_argument('--method', required=True, help='Fully qualified package and method to execute')
		parser_execute.add_argument('--data', required=True, help='Data payload')


		########## Discovery ###########
		parser_discovery = parser_sp1.add_parser('discover', help='Discover BP server on local subnet')
		parser_discovery.add_argument('--name', required=True, help='BP server containig specified key in service registry will respond')


		########## Email ###########


		########## Poll Account ###########


		########## Global ###########
		parser.add_argument('--bpbroker', '-b', metavar='host:port', help='BP Broker to communicate with')
		parser.add_argument('--access-key', '-a', default='', metavar='KEY', help='Optional access key required to access some BP Broker modules')
		parser.add_argument('--cols', nargs='*', metavar='COL', help='Include only specific columns in the output')
		parser.add_argument('--format', '-f', choices=['json','text','csv','csv-noheader'], default='text', help='Output result format (text is default)')
		self.args = parser.parse_args()


	def GetCommand(self):  return(self.args.command)
	def GetArgs(self):  return(self.args)


	def ImportIni(self):
		pass


	def MergeEnvironment(self):
		if 'BPBROKER' in os.environ:  bpclient.BPROKER = os.environ['BPROKER']


	def MergeCommands(self):
		if self.args.bpbroker:  bpclient.BPBROKER = self.args.bpbroker




class ExecCommand():
	def __init__(self):
		try:
			self.Bootstrap()
		except Exception as e:
			sys.exit(1)


	def Bootstrap(self):
		if bpclient.args.GetCommand() == 'ping':  self.Ping()
		elif bpclient.args.GetCommand() == 'discover':  self.Discover()
		elif bpclient.args.GetCommand() == 'execute':  self.Execute()
		elif bpclient.args.GetCommand() == 'service':  self.Services()


	def Services(self):
		if bpclient.args.GetArgs().sub_command == 'register':  self.ServicesRegister()
		elif bpclient.args.GetArgs().sub_command == 'get':  self.ServicesGet()
		elif bpclient.args.GetArgs().sub_command == 'replace':  self.ServicesReplace()
		elif bpclient.args.GetArgs().sub_command == 'update':  self.ServicesUpdate()
		elif bpclient.args.GetArgs().sub_command == 'delete':  self.ServicesDelete()


	def _ServicesWrapper(self,function,args,cols,opts={}):
		try:
			opts['supress_output'] = True
			r = self.Exec(function,args,cols,opts,supress_output=True)

			if 'success' in r and not r['success']:  raise(Exception(r['message']+"\n"))
			if 'data' in r and '_str' in r['data']:  r['data'] = r['data']['_str']
			if not bpclient.args.args.raw and 'data' in r: 
				r = {'data': r['data']}
				cols = ('data',)

			if 'data' in r:
				if bpclient.args.args.format == 'json':  print bpclient.output.Json(r,cols,opts)
				elif bpclient.args.args.format == 'text':  print bpclient.output.Text(r,cols,opts)
				elif bpclient.args.args.format == 'csv-noheader':  print bpclient.output.Csv(r,cols,{'no_header': True})
				elif bpclient.args.args.format == 'csv':  print bpclient.output.Csv(r,cols,opts)

		except Exception as e:
			sys.stderr.write("Fatal error: %s" % str(e))
			sys.exit(1)


	def Ping(self):
		self._ServicesWrapper('bpclient.ping.Ping',{'data': bpclient.args.args.data, 'access_key': bpclient.args.args.access_key},['src','pong'])


	def Execute(self):
		try:
			print self.Exec('bpclient.execute.Execute',
			                {'method': bpclient.args.args.method, 'data': bpclient.args.args.data, 'access_key': bpclient.args.args.access_key}, 
							[], supress_output=True)
		except Exception as e:
			sys.stderr.write("Fatal error: %s" % str(e))
			sys.exit(1)


	def Discover(self):
		try:
			self.Exec('bpclient.discover.Discover',{'name': bpclient.args.args.name}, ['bpbroker'])
		except Exception as e:
			sys.stderr.write("Fatal error: %s" % str(e))
			sys.exit(1)


	def ServicesRegister(self):
		self._ServicesWrapper('bpclient.services.Register',
		                      {'name': bpclient.args.args.name, 'data': bpclient.args.args.data, 'access_key': bpclient.args.args.access_key},
		                      ['success','message','data'])


	def ServicesGet(self):
		self._ServicesWrapper('bpclient.services.Get',
		                      {'name': bpclient.args.args.name, 'access_key': bpclient.args.args.access_key}, 
							  ['success','message','data'])


	def ServicesDelete(self):
		self._ServicesWrapper('bpclient.services.Delete',
		                      {'name': bpclient.args.args.name, 'access_key': bpclient.args.args.access_key}, 
							  ['success','message'])


	def ServicesReplace(self):
		self._ServicesWrapper('bpclient.services.Replace',
		                      {'name': bpclient.args.args.name, 'access_key': bpclient.args.args.access_key, 'data': bpclient.args.args.data},
		                      ['success','message','data'])


	def ServicesUpdate(self):
		self._ServicesWrapper('bpclient.services.Update',
		                      {'name': bpclient.args.args.name, 'data': bpclient.args.args.data, 'access_key': bpclient.args.args.access_key},
		                      ['success','message','data'])


	def Exec(self,function,args=False,cols=None,opts={},supress_output=False):
		try:
			if args:  r = eval("%s(**%s)" % (function,args))
			else:  r = eval("%s()" % (function))

			if bpclient.args.args.cols:  cols = bpclient.args.args.cols

			# TODO - not sure whether we'll reuse this code at all or duplicate in wrappers as needed?
			if not supress_output and bpclient.args.args.format == 'json':  print bpclient.output.Json(r,cols,opts)
			elif not supress_output and bpclient.args.args.format == 'text':  print bpclient.output.Text(r,cols,opts)
			elif not supress_output and bpclient.args.args.format == 'csv-noheader':  print bpclient.output.Csv(r,cols,{'no_header': True})
			elif not supress_output and bpclient.args.args.format == 'csv':  print bpclient.output.Csv(r,cols,opts)

			return(r)

		except:
			raise





"""
bp_broker ping module.

Basic package to test end to end communication
"""


import json

import bpbroker


#####################################################

def Ping(rh):
	"""Echo source host and querystring back in response.

	:param *: No params required
	:returns src: Requesting IP address
	:returns data: Echoes entire provided querystring
	"""

	data = None
	try:
		data = json.loads(rh.qs['data'])
	except:
		data = {'_str': rh.qs['data']}
	rh.data = json.dumps({'src': rh.RequestingHost(), 'data': data})





"""
bp_client execute module.

Package to initiate custom RPC call on BP Broker server
"""


import re
import requests

import bpclient


#####################################################

def Execute(method,data,access_key=''):
	"""Execute custom RPC on BP Broker server passing along data payload.

	Expects RPC method to be fully qualified incusive of file/package to execute and the actual method.
	For example if you were to execute the following on the BP Broker server:

		import testpacakage
		testpackage.testmodule.method(params)

	The equivalent call to this method is:

		bpclient.Execute(method="testpackage.testmodule.method",data="params")

	The actual call to the BP Broker is:

		POST https://broker/testpackage.testmodule/method?data=params

	If using the bpclient cli this is accomplished with the following command:

		bpclient --method testpackage.testmodule.method --data params
		
	:param method: RPC method to execute
	:param data: Data to be sent/echoed
	:returns : Sends returns all output to stdout
	"""

	method_match = re.match("(.*)\.(.*)",method)

	r = requests.post("https://%s/%s/%s/" % (bpclient.BPBROKER,method_match.group(1),method_match.group(2)),
	                  params={'access_key':access_key,'data': data},verify=False)
	return(r.text)





"""
bp_client discover module.

Package to initiate UDP broadcast to local networks searching for existing BP Broker servers.
"""


import socket
from socket import SOL_SOCKET, SO_BROADCAST
import sys

import bpclient


#####################################################

def Discover(name):
	"""Broadcast BP Broker discover traffic to local networks, returing BP Broker that resonds.

	Broadcasts UDP packet to all local networks containing a BP Broker discover query.
	If BP Broker is on one of these networks and contains the specified "name" in the service
	registry then it will respond.

	:param name: Services key to search to broadcast to bp broker servers
	:returns bpboker: Returns IP address of responding BP Broker or False if no response
	"""

	# SOCK_DGRAM is the socket type to use for UDP sockets
	sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	sock.settimeout(3)
	sock.setsockopt(SOL_SOCKET, SO_BROADCAST, True)

	try:
		sock.sendto(name + "\n", ('<broadcast>', 20443))
		received, src = sock.recvfrom(1024)

		if received == "Key Exists\n":  return({'bpbroker': src[0]})
		else:  raise(Exception("No results"))

	except socket.timeout:
		raise(Exception("No results"))





#
# python setup.py sdist
# python setup.py bdist_dumb
# python setup.py bdist_rpm
#

from setuptools import setup, find_packages

setup(
	name = "bpbroker",
	version = "0.11",
	packages = find_packages("."),

	package_data = {
		"bpbroker": ["dummy_api.crt","dummy_api.key"],
	},

	install_requires = ['argparse','requests'],

	entry_points = {
		'console_scripts': [
			'bpbroker  = bpbroker.cli:BPBroker',
			'bpclient  = bpclient.cli:BPClient',
			'bpmailer  = bpclient.cli:BPMailer',
		],
	},


	# metadata for upload to PyPI
	author = "Keith Resar",
	author_email = "Keith.Resar@CenturyLinkCloud.com",
	description = "CenturyLink Cloud Blueprint Broker Service",
	keywords = "CenturyLink Cloud Blueprint Broker",
	url = "https://github.com/CenturyLinkCloud/Ecosystem/tree/master/Blueprints/Blueprint%20Broker",

	# could also include long_description, download_url, classifiers, etc.
)

